# Architecture of Explorer's frontend

## Overview

The frontend of Explorer has been written in [PureScript](http://www.purescript.org/). Its architecture is based on [_The Elm Architecture (TEA)_](https://guide.elm-lang.org/architecture/) by using [Pux](http://purescript-pux.org/). With that the application logic of Explorer is break up into three cleanly separated parts:

* **Model** - _State of application_   
* **Update** - _Updates the application state_
* **View** - _Displays the application state as HTML_

### Model

In module [`Explorer.Types.State`](src/Explorer/Types/State.purs) are all types of Explorers application state defined. By [starting the application](#main) a [`initialState`](src/Explorer/State.purs) is set. All values of the monolithic state are accessible by using `Lens`, which [are generated by building](README.md#2-generate-lenses) Explorer.

### Update

While running the application every `Action` defined in [`Explorer.Types.Actions`](src/Explorer/Types/Actions.purs) (e.g. an interaction of an user by clicking a button to paginate a list of `epochs`) ends up in [`Explorer.Update`](src/Explorer/Update.purs) to update the application `State` or to do some `IO` effects, such as requesting data of a [`transaction detail page`](https://cardanoexplorer.com/tx/d748269f80af5d565c7b7bf62f248db4de7ae3f502f0205dee7f1016c0e7fb35).

### View

All views are rendered into HTML and are defined in folder [`src/Explorer/View`](src/Explorer/View). The top most located view of the HTML hierarchy is defined in [`Explorer.View.Layout`](explorer/frontend/src/Explorer/View/Layout.purs), which wraps all other, nested child-views.

In module [`Explorer.View.Common`](explorer/frontend/src/Explorer/View/Common.purs) are all shared views defined, for example `txPaginationView` to paginate a list of `epochs` at the [`dashboard`](https://cardanoexplorer.com/) or `transactions` at the [`address detail page`](https://cardanoexplorer.com/address/DdzFFzCqrhsoDF375UrpTZdTEpuADS5ktEWSgaimqt6GGwdD8YUQUEZSugciZzzNJky3MfGoodsfyebzxvj9BGcfryieojff69vP4zUS).

Every interaction of an user (e.g. clicking `expand` / `collapse` button of `transaction feed` list) fires an [`Action`](src/Explorer/Types/Actions.purs), which will be handled in [`Explorer.Update`](src/Explorer/Update.purs) (_as described in section [`Update`](#update) above_).

## Other parts

### Main

The `main` entry point of the application is defined in [`Main`](src/Main.purs) module. By starting the application all [`Action`](src/Explorer/Types/Actions.purs)'s, [`Routes`](src/Explorer/Routes.purs), [i18n](Explorer.I18n.Types) etc. are registered and configurations values (e.g. running in `testnet`) are set.

### Routes

Explorer is running as a [_Single Page Application (SPA)_](https://en.wikipedia.org/wiki/Single-page_application). All routes are defined in [`Explorer.Routes`](src/Explorer/Routes.purs). Deep linking is enabled by default, which means that every sub-page, such as a [`address detail page`](https://cardanoexplorer.com/address/DdzFFzCqrhsoDF375UrpTZdTEpuADS5ktEWSgaimqt6GGwdD8YUQUEZSugciZzzNJky3MfGoodsfyebzxvj9BGcfryieojff69vP4zUS), can be reached directly by loading all necessary data independently with different routes.

### Sockets

Explorer uses [`sockets-io`](https://socket.io) to get data in real-time. All socket event handlers are defined in [`Explorer.Api.Socket`](explorer/frontend/src/Explorer/Api/Socket.purs). Sockets events are registered like any other `Actions` in [`Main.purs`](src/Main.purs) as described in section [`Main`](#main).

### HTTP

All handlers for requesting _no real-time_ data by HTTP are defined in module [`Explorer.Api.Http`](explorer/frontend/src/Explorer/Api/Http.purs)


### Lib

In folder [`Lib`](src/Lib/) you will find custom PureScript bindings for using _native_ JavaScript libraries, such as [`Waypoints`](src/Lib/Waypoints/Waypoints.purs) and [`BigNumber`](src/Lib/BigNumber/BigNumber.purs).

### i18n

Explorer is an `i18n` application. All internationalization content are located in folder [`src/Explorer/I18n`](src/Explorer/I18n), such as translation of [`English`](Explorer/I18n/EN.purs) and [`Japanese`](src/Explorer/I18n/JP.purs).


### Utilities

_Not surprising_, we do have couple of helpers. Those are defined in folder [`src/Explorer/I18n`](src/Explorer/I18n), such as for handling [`Strings`](src/Explorer/Util/String.purs), [`Time`](src/Explorer/Util/Time.purs) or [`DOM`](src/Explorer/Util/DOM.purs).
